@@ -22,7 +22,6 @@
 
 
 List	   *getOptions(Oid foreigntableid);
-PyObject   *optionsListToPyDict(List *options);
 bool		compareOptions(List *options1, List *options2);
 
 void		getColumnsFromTable(TupleDesc desc, PyObject **p_columns, List **columns);
@@ -37,7 +36,6 @@ PyObject *pythonQual(char *operatorname, PyObject *value,
 		   bool use_or,
 		   Oid typeoid);
 
-const char *getPythonEncodingName(void);
 
 
 
@@ -55,7 +53,7 @@ PyObject   *datumNumberToPython(Datum node, ConversionInfo * cinfo);
 PyObject   *datumDateToPython(Datum datum, ConversionInfo * cinfo);
 PyObject   *datumTimestampToPython(Datum datum, ConversionInfo * cinfo);
 PyObject   *datumIntToPython(Datum datum, ConversionInfo * cinfo);
-PyObject   *datumArrayToPython(Datum datum, ConversionInfo * cinfo);
+PyObject   *datumArrayToPython(Datum datum, Oid type, ConversionInfo * cinfo);
 PyObject   *datumByteaToPython(Datum datum, ConversionInfo * cinfo);
 PyObject   *datumUnknownToPython(Datum datum, ConversionInfo * cinfo, Oid type);
 
@@ -95,11 +93,6 @@ void appendBinaryStringInfoQuote(StringInfo buffer,
 							Py_ssize_t strlength,
 							bool need_quote);
 
-UserMapping *
-			multicorn_GetUserMapping(Oid userid, Oid serverid);
-
-
-
 
 static void begin_remote_xact(CacheEntry * entry);
 
@@ -132,7 +125,7 @@ PyUnicode_AsPgString(PyObject *p_unicode)
 	unicode_size = PyUnicode_GET_SIZE(p_unicode);
 	pTempStr = PyUnicode_Encode(PyUnicode_AsUnicode(p_unicode),
 								unicode_size,
-								GetDatabaseEncodingName(), NULL);
+								getPythonEncodingName(), NULL);
 	errorCheck();
 	message = strdup(PyBytes_AsString(pTempStr));
 	errorCheck();
@@ -413,6 +406,7 @@ optionsListToPyDict(List *options)
 	return p_options_dict;
 }
 
+
 bool
 compareOptions(List *options1, List *options2)
 {
@@ -768,14 +762,7 @@ qualdefToPython(MulticornConstQual * qualdef, ConversionInfo ** cinfos)
 		{
 			typeoid = cinfo->atttypoid;
 		}
-		if (qualdef->base.isArray)
-		{
-			p_value = datumArrayToPython(value, cinfo);
-		}
-		else
-		{
-			p_value = datumToPython(value, typeoid, cinfo);
-		}
+		p_value = datumToPython(value, typeoid, cinfo);
 		if (p_value == NULL)
 		{
 			return NULL;
@@ -1347,10 +1334,16 @@ datumIntToPython(Datum datum, ConversionInfo * cinfo)
 }
 
 PyObject *
-datumArrayToPython(Datum datum, ConversionInfo * cinfo)
+datumArrayToPython(Datum datum, Oid type, ConversionInfo * cinfo)
 {
+#if PG_VERSION_NUM >= 90500
+	ArrayIterator iterator = array_create_iterator(DatumGetArrayTypeP(datum),
+												   0, NULL);
+# else
 	ArrayIterator iterator = array_create_iterator(DatumGetArrayTypeP(datum),
 												   0);
+# endif
+
 	Datum		elem = (Datum) NULL;
 	bool		isnull;
 	PyObject   *result = PyList_New(0),
@@ -1364,7 +1357,19 @@ datumArrayToPython(Datum datum, ConversionInfo * cinfo)
 		}
 		else
 		{
-			pyitem = datumToPython(elem, cinfo->atttypoid, cinfo);
+			HeapTuple	tuple;
+			Form_pg_type typeStruct;
+
+			tuple = SearchSysCache1(TYPEOID, ObjectIdGetDatum(type));
+			if (!HeapTupleIsValid(tuple))
+			{
+				elog(ERROR, "lookup failed for type %u",
+					 type);
+			}
+			typeStruct = (Form_pg_type) GETSTRUCT(tuple);
+			pyitem = datumToPython(elem, typeStruct->typelem, cinfo);
+			ReleaseSysCache(tuple);
+
 			PyList_Append(result, pyitem);
 			Py_DECREF(pyitem);
 		}
@@ -1422,7 +1427,7 @@ datumToPython(Datum datum, Oid type, ConversionInfo * cinfo)
 			if ((typeStruct->typelem != 0) && (typeStruct->typlen == -1))
 			{
 				/* Its an array. */
-				return datumArrayToPython(datum, cinfo);
+				return datumArrayToPython(datum, type, cinfo);
 			}
 			return datumUnknownToPython(datum, cinfo, type);
 	}
@@ -1506,7 +1511,7 @@ tupleTableSlotToPyObject(TupleTableSlot *slot, ConversionInfo ** cinfos)
 		PyObject   *item;
 		AttrNumber	cinfo_idx = attr->attnum - 1;
 
-		if (attr->attisdropped)
+		if (attr->attisdropped || cinfos[cinfo_idx] == NULL)
 		{
 			continue;
 		}
